"""
Defines the interface for the "forward" task,
which can be used for recognition, alignment, search, etc.

https://github.com/rwth-i6/returnn/issues/1336
"""

from __future__ import annotations
from returnn.tensor import TensorDict


class ForwardCallbackIface:
    """
    Callback interface for the forward task.

    Define `forward_callback` in your config to an instance or class of this.

    https://github.com/rwth-i6/returnn/issues/1336
    """

    def init(self, *, model):
        """
        Run at the beginning.
        """

    def process_seq(self, *, seq_tag: str, outputs: TensorDict):
        """
        Called for each sequence, or entry in the dataset.
        This does not have the batch dim anymore.
        The values in `outputs` are Numpy arrays.

        :param seq_tag:
        :param outputs:
        """

    def finish(self):
        """
        Run at the end.
        """



#!/usr/bin/env python3

"""
Benchmarking LSTM training times for different LSTM implementations, both on GPU and CPU.

Some runtimes:

GeForce GTX 680:
  Settings:
  {'batch_size': 2000,
   'chunking': '50:25',
   'max_seqs': 40,
   'n_hidden': 500,
   'num_layers': 5,
   'num_seqs': 500}
  Final results:
    GPU:CudnnLSTM: 0:00:12.3727
    GPU:NativeLSTM: 0:00:19.0306
    GPU:LSTMBlockFused: 0:00:41.5129
    GPU:LSTMBlock: 0:00:52.6296
    GPU:StandardLSTM: 0:00:53.2629
    GPU:BasicLSTM: 0:00:53.6150
    CPU:NativeLSTM: 0:02:37.1753
    CPU:LSTMBlockFused: 0:03:17.5448
    CPU:BasicLSTM: 0:03:28.9882
    CPU:StandardLSTM: 0:03:29.3730
    CPU:LSTMBlock: 0:03:33.7039

  Settings:
  {'batch_size': 2000,
   'chunking': '0',
   'max_seqs': 50,
   'n_hidden': 500,
   'num_layers': 5,
   'num_seqs': 500}
  Final results:
    GPU:CudnnLSTM: 0:00:12.8325
    GPU:NativeLSTM: 0:00:15.8518
    GPU:LSTMBlockFused: 0:01:05.2157
    GPU:StandardLSTM: 0:01:20.6564
    GPU:BasicLSTM: 0:01:20.9896
    GPU:LSTMBlock: 0:01:21.4046
    CPU:NativeLSTM: 0:02:12.5029
    CPU:LSTMBlockFused: 0:03:39.0270
    CPU:LSTMBlock: 0:03:51.9667
    CPU:StandardLSTM: 0:03:56.6404
    CPU:BasicLSTM: 0:03:58.1545
"""

from __future__ import annotations
import sys
import time
from argparse import ArgumentParser
from pprint import pprint

import _setup_returnn_env  # noqa
from returnn.util import better_exchook
from returnn.log import log
from returnn.config import Config
from returnn.util.basic import hms_fraction, describe_returnn_version, describe_tensorflow_version
from returnn.tf.engine import Engine
from returnn.tf.util.basic import setup_tf_thread_pools, is_gpu_available, print_available_devices
from returnn.datasets import init_dataset, Dataset


LstmCellTypes = [
    "BasicLSTM",
    "StandardLSTM",
    "LSTMBlock",
    "LSTMBlockFused",
    "NativeLSTM",
    "NativeLstm2",
    "NativeLstmLowMem",
    "CudnnLSTM",
]

GpuOnlyCellTypes = ["CudnnLSTM"]

# Fixed by dataset. See make_config_dict().
_input_dim = 9
_output_dim = 2

# You can play around with these. E.g. use "chunking=0", "max_seqs=20" as command-line args.
base_settings = {
    "num_layers": 5,  # number of LSTM layers
    "n_hidden": 500,  # per direction, per layer, the number of units
    "num_seqs": 500,  # for the dataset generation
    "batch_size": 2000,  # upper limit for n_seqs * max_seq_len in a batch
    "max_seqs": 40,  # upper limit for n_seqs in a batch
    "chunking": "50:25",  # set to "0" to disable
}


def make_config_dict(lstm_unit, use_gpu):
    """
    :param str lstm_unit: "NativeLSTM", "LSTMBlock", "LSTMBlockFused", "CudnnLSTM", etc, one of LstmCellTypes
    :param bool use_gpu:
    :return: config dict
    :rtype: dict[str]
    """
    num_layers = base_settings["num_layers"]
    network = {}
    for i in range(num_layers):
        for direction in [-1, 1]:
            dir_str = {-1: "bwd", 1: "fwd"}[direction]
            layer = {"class": "rec", "unit": lstm_unit, "n_out": base_settings["n_hidden"], "direction": direction}
            if i > 0:
                layer["from"] = ["lstm%i_fwd" % i, "lstm%i_bwd" % i]
            network["lstm%i_%s" % (i + 1, dir_str)] = layer
    network["output"] = {
        "class": "softmax",
        "loss": "ce",
        "target": "classes",
        "from": ["lstm%i_fwd" % num_layers, "lstm%i_bwd" % num_layers],
    }
    return {
        "device": "gpu" if use_gpu else "cpu",
        "train": {"class": "Task12AXDataset", "num_seqs": base_settings["num_seqs"]},
        "num_inputs": _input_dim,
        "num_outputs": _output_dim,
        "num_epochs": 1,
        "model": None,  # don't save
        "tf_log_dir": None,  # no TF logs
        "tf_log_memory_usage": True,
        "network": network,
        # batching
        "batch_size": base_settings["batch_size"],
        "max_seqs": base_settings["max_seqs"],
        "chunking": base_settings["chunking"],
        # optimization
        "optimizer": {"class": "adam"},
        "learning_rate": 0.01,
    }


def benchmark(lstm_unit, use_gpu):
    """
    :param str lstm_unit: e.g. "LSTMBlock", one of LstmCellTypes
    :param bool use_gpu:
    :return: runtime in seconds of the training itself, excluding initialization
    :rtype: float
    """
    device = {True: "GPU", False: "CPU"}[use_gpu]
    key = "%s:%s" % (device, lstm_unit)
    print(">>> Start benchmark for %s." % key)
    config = Config()
    config.update(make_config_dict(lstm_unit=lstm_unit, use_gpu=use_gpu))
    dataset_kwargs = config.typed_value("train")
    Dataset.kwargs_update_from_config(config, dataset_kwargs)
    dataset = init_dataset(dataset_kwargs)
    engine = Engine(config=config)
    engine.init_train_from_config(config=config, train_data=dataset)
    print(">>> Start training now for %s." % key)
    start_time = time.time()
    engine.train()
    runtime = time.time() - start_time
    print(">>> Runtime of %s: %s" % (key, hms_fraction(runtime)))
    engine.finalize()
    return runtime


def main():
    """
    Main entry.
    """
    global LstmCellTypes
    print("Benchmarking LSTMs.")
    better_exchook.install()
    print("Args:", " ".join(sys.argv))
    arg_parser = ArgumentParser()
    arg_parser.add_argument("cfg", nargs="*", help="opt=value, opt in %r" % sorted(base_settings.keys()))
    arg_parser.add_argument("--no-cpu", action="store_true")
    arg_parser.add_argument("--no-gpu", action="store_true")
    arg_parser.add_argument("--selected", help="comma-separated list from %r" % LstmCellTypes)
    arg_parser.add_argument("--no-setup-tf-thread-pools", action="store_true")
    args = arg_parser.parse_args()
    for opt in args.cfg:
        key, value = opt.split("=", 1)
        assert key in base_settings
        value_type = type(base_settings[key])
        base_settings[key] = value_type(value)
    print("Settings:")
    pprint(base_settings)

    log.initialize(verbosity=[4])
    print("Returnn:", describe_returnn_version(), file=log.v3)
    print("TensorFlow:", describe_tensorflow_version(), file=log.v3)
    print("Python:", sys.version.replace("\n", ""), sys.platform)
    if not args.no_setup_tf_thread_pools:
        setup_tf_thread_pools(log_file=log.v2)
    else:
        print("Not setting up the TF thread pools. Will be done automatically by TF to number of CPU cores.")
    if args.no_gpu:
        print("GPU will not be used.")
    else:
        print("GPU available: %r" % is_gpu_available())
    print_available_devices()

    if args.selected:
        LstmCellTypes = args.selected.split(",")
    benchmarks = {}
    if not args.no_gpu and is_gpu_available():
        for lstm_unit in LstmCellTypes:
            benchmarks["GPU:" + lstm_unit] = benchmark(lstm_unit=lstm_unit, use_gpu=True)
    if not args.no_cpu:
        for lstm_unit in LstmCellTypes:
            if lstm_unit in GpuOnlyCellTypes:
                continue
            benchmarks["CPU:" + lstm_unit] = benchmark(lstm_unit=lstm_unit, use_gpu=False)

    print("-" * 20)
    print("Settings:")
    pprint(base_settings)
    print("Final results:")
    for t, lstm_unit in sorted([(t, lstm_unit) for (lstm_unit, t) in sorted(benchmarks.items())]):
        print("  %s: %s" % (lstm_unit, hms_fraction(t)))
    print("Done.")


if __name__ == "__main__":
    main()



import sys
import _setup_test_env  # noqa
import unittest
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_greater, assert_true, assert_false
from pprint import pprint
from returnn.config import *
from returnn.util import better_exchook
from io import StringIO
import textwrap


def test_old_format():
    config = Config()
    config.load_file(
        StringIO(
            """
  # comment
  num_inputs 3
  hidden_type forward,lstm
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_false(config.is_typed("num_inputs"))
    assert_false(config.is_typed("hidden_type"))


def test_json_format():
    config = Config()
    config.load_file(
        StringIO(
            """
  {
  // comment
  "num_inputs": 3,
  "hidden_type": ["forward", "lstm"]
  }
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_true(config.is_typed("num_inputs"))
    assert_true(config.is_typed("hidden_type"))
    assert_is_instance(config.typed_value("num_inputs"), int)
    assert_is_instance(config.typed_value("hidden_type"), list)
    assert_equal(config.typed_value("hidden_type"), ["forward", "lstm"])


def test_py_config():
    config = Config()
    config.load_file(
        StringIO(
            """#!rnn.py
# comment
num_inputs = 3
hidden_type = ["forward", "lstm"]
  """
        )
    )

    assert_true(config.has("num_inputs"))
    assert_true(config.has("hidden_type"))
    assert_equal(config.int("num_inputs", -1), 3)
    assert_equal(config.value("hidden_type", "x"), "forward,lstm")
    assert_equal(config.value("hidden_type", "x", index=0), "forward")
    assert_equal(config.value("hidden_type", "x", index=1), "lstm")
    assert_equal(config.list("hidden_type", ["x"]), ["forward", "lstm"])

    assert_true(config.is_typed("num_inputs"))
    assert_true(config.is_typed("hidden_type"))
    assert_is_instance(config.typed_value("num_inputs"), int)
    assert_is_instance(config.typed_value("hidden_type"), list)
    assert_equal(config.typed_value("hidden_type"), ["forward", "lstm"])


def test_rnn_init_config_py_global_var():
    import returnn.__main__ as rnn
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".config", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """#!rnn.py
task = config.value("task", "train")

test_value = 42

def test_func():
  return task

    """
        )
        cfgfile.flush()
        rnn.init_config(command_line_options=[cfgfile.name, "--task", "search"])

    assert isinstance(rnn.config, Config)
    rnn.config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(rnn.config.dict)
    pprint(rnn.config.typed_dict)
    assert rnn.config.has("task")
    assert rnn.config.has("test_value")
    assert rnn.config.has("test_func")
    assert_equal(rnn.config.value("task", None), "search")
    assert rnn.config.is_typed("test_value")
    assert_equal(rnn.config.typed_value("test_value"), 42)
    assert rnn.config.is_typed("test_func")
    # So far it's fine.
    # Now something a bit strange.
    # Earlier, this failed, because the command-line overwrote this.
    assert rnn.config.is_typed("task")
    test_func = rnn.config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), "search")


def test_rnn_init_config_py_cmd_type():
    import returnn.__main__ as rnn
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".config", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """#!rnn.py
max_seq_length = {'bpe': 75}

def test_func():
  return max_seq_length

    """
        )
        cfgfile.flush()
        rnn.init_config(command_line_options=[cfgfile.name, "++max_seq_length", "0"])

    assert isinstance(rnn.config, Config)
    assert rnn.config.has("max_seq_length")
    assert rnn.config.has("test_func")
    assert rnn.config.is_typed("max_seq_length")
    assert rnn.config.is_typed("test_func")
    test_func = rnn.config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), 0)


def test_config_py_ext():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """
def test_func():
  return config.value("task", "train")
    """
        )
        cfgfile.flush()
        config = Config()
        config.load_file(cfgfile.name)  # should determine format by suffix ".py"

    config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(config.dict)
    pprint(config.typed_dict)
    assert config.has("test_func")
    assert config.is_typed("test_func")
    test_func = config.typed_dict["test_func"]
    assert callable(test_func)
    assert_equal(test_func(), "train")
    config.set("task", "search")
    assert_equal(test_func(), "search")


def test_config_py_old_returnn_imports():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", prefix="test_rnn_initConfig") as cfgfile:
        cfgfile.write(
            """
# These are some common imports found in older config files.
from Pretrain import WrapEpochValue
from TFUtil import where_bc
import TFUtil
import returnn.TFUtil

import TFHorovod
hvd = TFHorovod.get_ctx(config=config)  # should return None if no Horovod context

from returnn.Util import describe_returnn_version
returnn_version = describe_returnn_version()

tf_version_tuple = returnn.TFUtil.tf_version_tuple()

    """
        )
        cfgfile.flush()
        config = Config()
        config.load_file(cfgfile.name)  # should determine format by suffix ".py"

    config.typed_dict.pop("__builtins__", None)  # not needed, too verbose for pprint
    pprint(config.typed_dict)

    import returnn.util.basic as util
    import returnn.tf.util.basic as tf_util

    assert config.typed_dict["where_bc"] is tf_util.where_bc
    assert config.typed_dict["TFUtil"].where_bc is tf_util.where_bc
    assert config.typed_dict["hvd"] is None
    assert config.typed_dict["tf_version_tuple"] == tf_util.tf_version_tuple()
    assert config.typed_dict["returnn_version"] == util.describe_returnn_version()
    assert config.typed_dict["returnn"].TFUtil is tf_util


def test_pickle_config():
    import pickle
    import io

    config = Config()
    config.load_file(
        StringIO(
            textwrap.dedent(
                """\
                #!returnn.py

                def my_custom_func():
                    return 42

                class CustomClass:
                    x = 43

                    def __init__(self):
                        super().__init__()
                        CustomClass.x = 44

                    def get_value(self):
                        return CustomClass.x
                """
            )
        )
    )
    f = config.typed_dict["my_custom_func"]
    obj = config.typed_dict["CustomClass"]()

    sio = io.BytesIO()
    # noinspection PyUnresolvedReferences
    p = pickle._Pickler(sio)  # better for debugging
    with global_config_ctx(config):
        p.dump(config)

    config_ = pickle.loads(sio.getvalue())
    f_ = config_.typed_dict["my_custom_func"]
    assert f is not f_  # it should really be a copy, via Config.__getstate__ logic
    assert f() == f_() == 42
    obj_ = config_.typed_dict["CustomClass"]()
    assert type(obj) is not type(obj_)
    assert obj.get_value() == obj_.get_value() == 44


def test_config_pickle_function():
    # Having some function inside the config, there are cases when we need to pickle it,
    # e.g. when it is the post_process function of a dataset,
    # and then used as the PyTorch dataset -- see test_PTDataset.py, test_HDFDataset_pickle.
    # Currently, it fails with:
    #   PicklingError: Can't pickle <function speed_pert_...>:
    #     import of module '__returnn_config__' failed
    # Functions are pickled by reference, storing the module name and function name.
    import pickle

    config = Config()
    config.load_file(
        StringIO(
            textwrap.dedent(
                """\
                #!returnn.py

                def my_custom_func():
                    return 42
                """
            )
        )
    )
    with global_config_ctx(config):
        f = config.typed_dict["my_custom_func"]
        f_ = pickle.loads(pickle.dumps(f))
        assert f_ is f
        assert f_() == 42


def _config_pickle_proc_main(config, f):
    assert isinstance(config, Config)
    assert get_global_config() is config
    assert callable(f)
    f()


def test_config_pickle_function_multi_proc():
    # Same as test_config_pickle_function but via multiprocessing,
    # so across process boundaries.
    import multiprocessing

    _mp = multiprocessing.get_context("spawn")

    config = Config()
    config.load_file(
        StringIO(
            textwrap.dedent(
                """\
                #!returnn.py

                def my_custom_func():
                    import sys
                    sys.exit(42)
                """
            )
        )
    )
    with global_config_ctx(config):
        f = config.typed_dict["my_custom_func"]
        proc = _mp.Process(target=_config_pickle_proc_main, args=(config, f))
        proc.start()
        proc.join()
        assert proc.exitcode == 42


def test_config_pickle_twice():
    import pickle
    import io

    config = Config()
    config.load_file(
        StringIO(
            textwrap.dedent(
                """\
                #!returnn.py

                def my_custom_func():
                    return 42
                """
            )
        )
    )

    def _pickle(config_: Config) -> bytes:
        sio = io.BytesIO()
        # noinspection PyUnresolvedReferences
        p = pickle._Pickler(sio)  # better for debugging
        with global_config_ctx(config_):
            p.dump(config_)
        return sio.getvalue()

    def _unpickle(data_: bytes) -> Config:
        with global_config_ctx(None):  # guard orig global config setting
            return pickle.loads(data_)

    config = _unpickle(_pickle(config))
    config = _unpickle(_pickle(config))

    with global_config_ctx(config):
        f = config.typed_dict["my_custom_func"]
        f_ = pickle.loads(pickle.dumps(f))
        assert f_ is f
        assert f_() == 42


def test_config_pickle_function_multi_proc_twice():
    # Same as test_config_pickle_function but via multiprocessing,
    # so across process boundaries.
    import multiprocessing

    _mp = multiprocessing.get_context("spawn")

    config = Config()
    config.load_file(
        StringIO(
            textwrap.dedent(
                """\
                #!returnn.py

                def my_custom_func():
                    import sys
                    sys.exit(42)
                """
            )
        )
    )
    with global_config_ctx(config):
        f = config.typed_dict["my_custom_func"]
        proc = _mp.Process(target=_config_pickle_twice_proc_main, args=(config, f))
        proc.start()
        proc.join()
        assert proc.exitcode == 42


def _config_pickle_twice_proc_main(config, f):
    assert isinstance(config, Config)
    assert get_global_config() is config
    assert callable(f)

    import multiprocessing

    _mp = multiprocessing.get_context("spawn")
    proc = _mp.Process(target=_config_pickle_proc_main, args=(config, f))
    proc.start()
    proc.join()
    assert proc.exitcode == 42

    f()


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
prev tensor ref for loop, i.e. RecLayer, specifically "prev:..." layer references
"""

from __future__ import annotations
from returnn.tensor import Tensor
from .. import frontend_layers as rfl


__all__ = ["PrevTensorRef"]


class PrevTensorRef(Tensor):
    """
    Refers to a layer from the previous loop iteration.
    """

    @classmethod
    def get_prev_ref(cls, *, cur_layer_name_ctx: rfl.Layer, initial: Tensor) -> PrevTensorRef:
        """
        Create prev ref.
        """
        parent_name_ctx = cur_layer_name_ctx.parent
        prev_tensor_name_ctx = parent_name_ctx.get_child(f"prev:{cur_layer_name_ctx.name}")
        if prev_tensor_name_ctx.tensor:
            prev_tensor_ref = prev_tensor_name_ctx.tensor
            assert isinstance(prev_tensor_ref, PrevTensorRef)
            assert prev_tensor_ref.cur_layer_name_ctx is cur_layer_name_ctx
        else:
            prev_tensor_ref = PrevTensorRef(
                name_ctx=prev_tensor_name_ctx, cur_layer_name_ctx=cur_layer_name_ctx, data=initial
            )
            assert prev_tensor_name_ctx.tensor is prev_tensor_ref
        return prev_tensor_ref

    def __init__(self, *, name_ctx: rfl.Layer, cur_layer_name_ctx: rfl.Layer, data: Tensor):
        # At the time we instantiate this, cur_layer_name_ctx.tensor probably does not exist yet.
        super().__init__(**data.get_kwargs())
        name_ctx.tensor = self
        self.raw_tensor = name_ctx
        self.cur_layer_name_ctx = cur_layer_name_ctx
        self.raw_tensor.layer_extra_dependencies.append(self.cur_layer_name_ctx)

    def assign_new_cur_tensor_name_ctx(self, cur_tensor_name_ctx: rfl.Layer):
        """
        Changes self.name_ctx to new name_ctx.
        """
        self.raw_tensor.layer_extra_dependencies.remove(self.cur_layer_name_ctx)
        prev_layer_name = f"prev:{cur_tensor_name_ctx.name}"
        assert prev_layer_name not in cur_tensor_name_ctx.parent.children
        prev_layer_name_ctx = cur_tensor_name_ctx.parent.get_child(prev_layer_name)
        prev_layer_name_ctx.move_tensor_here(self)
        assert self.raw_tensor is prev_layer_name_ctx
        self.cur_layer_name_ctx = cur_tensor_name_ctx
        self.raw_tensor.layer_extra_dependencies.append(self.cur_layer_name_ctx)



"""
Transformer encoder

Also see :mod:`returnn.frontend.decoder.transformer`.
"""

from __future__ import annotations
from typing import Optional, Union, Any, Callable, Sequence, Dict
import functools
import copy as _copy
from types import FunctionType
from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from returnn.frontend.decoder.transformer import make_norm, FeedForward


class TransformerEncoder(rf.Module):
    """
    Represents the Transformer encoder architecture
    """

    def __init__(
        self,
        vocab_dim: Dim,
        model_dim: Union[Dim, int] = Dim(512, name="transformer-enc-default-model-dim"),
        *,
        num_layers: int,
        ff: Union[type, Dict[str, Any], rf.Module] = NotSpecified,
        pos_enc: Union[None, Callable, Dict[str, Any], rf.Module] = rf.sinusoidal_positional_encoding,
        dropout: float = 0.1,
        num_heads: int = 8,
        att_dropout: float = 0.1,
        norm: Union[type, Dict[str, Any], rf.Module, Callable] = rf.LayerNorm,
        decoder_layer: Optional[Union[TransformerEncoderLayer, rf.Module, type, Any]] = None,
        embed_dim: Optional[Dim] = None,
        input_embedding_scale: float = None,
        input_dropout: float = None,
        sequential=rf.Sequential,
    ):
        """
        :param vocab_dim:
        :param model_dim: the output feature dimension
        :param num_layers: the number of encoder layers
        :param ff: feed-forward / MLP block. Default is :class:`FeedForward`
        :param pos_enc: positional encoding. Default is sinusoidal positional encoding.
        :param dropout: the dropout value for the FF block
        :param num_heads: the number of attention heads
        :param att_dropout: attention dropout value
        :param norm: pre-normalization for FF and attention blocks
        :param decoder_layer: an instance of :class:`TransformerDecoderLayer` or similar
        :param embed_dim: if given, will first have an embedding [vocab,embed] and then a linear [embed,model].
        :param input_embedding_scale:
        :param input_dropout:
        :param sequential:
        """
        super().__init__()

        if not isinstance(vocab_dim, Dim):
            raise TypeError(f"TransformerDecoder: unexpected vocab_dim {vocab_dim!r} type {type(vocab_dim)}")
        if isinstance(model_dim, int):
            model_dim = Dim(model_dim, name="transformer-dec-model-dim")
        if not isinstance(model_dim, Dim):
            raise TypeError(f"TransformerDecoder: unexpected model_dim {model_dim!r} type {type(model_dim)}")

        self.vocab_dim = vocab_dim
        self.model_dim = model_dim
        self.embed_dim = embed_dim

        # We could make this optional or configurable if we ever need to.
        # Or maybe you would just have another separate implementation of this module then...
        self.input_embedding = rf.Embedding(vocab_dim, embed_dim or model_dim)

        self.input_embedding_proj = None
        if embed_dim:
            self.input_embedding_proj = rf.Linear(embed_dim, model_dim, with_bias=False)

        if pos_enc is None:
            pass
        elif isinstance(pos_enc, dict):
            pos_enc = rf.build_from_dict(
                pos_enc, feat_dim=embed_dim or model_dim, dtype=self.input_embedding.weight.dtype
            )
        elif isinstance(pos_enc, rf.Module):
            pass
        elif isinstance(pos_enc, FunctionType):
            pos_enc = functools.partial(
                pos_enc, feat_dim=embed_dim or model_dim, dtype=self.input_embedding.weight.dtype
            )
        else:
            raise TypeError(f"unexpected pos_enc type {pos_enc!r}")
        self.pos_enc = pos_enc
        if input_embedding_scale is None:
            input_embedding_scale = model_dim.dimension**0.5
        self.input_embedding_scale = input_embedding_scale
        if input_dropout is None:
            input_dropout = dropout
        self.input_dropout = input_dropout

        if not decoder_layer or isinstance(decoder_layer, type):
            decoder_layer_opts_ = dict(
                out_dim=model_dim,
                ff=ff,
                dropout=dropout,
                num_heads=num_heads,
                att_dropout=att_dropout,
                norm=norm,
            )
            if not decoder_layer:
                decoder_layer = TransformerEncoderLayer(**decoder_layer_opts_)
            elif isinstance(decoder_layer, type):
                decoder_layer = decoder_layer(**decoder_layer_opts_)
            else:
                raise TypeError(f"unexpected decoder_layer {decoder_layer!r}")

        self.layers = sequential(_copy.deepcopy(decoder_layer) for _ in range(num_layers))

        self.final_layer_norm = make_norm(norm, model_dim)

    def default_initial_state(self, *, batch_dims: Sequence[Dim]) -> rf.State:
        """default initial state"""
        state = rf.State({k: v.default_initial_state(batch_dims=batch_dims) for k, v in self.layers.items()})
        state.pos = rf.zeros((), dtype="int32", device="cpu")
        return state

    def transform_encoder(self, encoder: Tensor, *, axis: Dim) -> rf.State:
        """
        Transform encoder output.
        Note that the Transformer decoder usually expects that layer-norm was applied already on the encoder output.
        """
        return rf.State({k: v.transform_encoder(encoder, axis=axis) for k, v in self.layers.items()})

    def __call__(
        self, source: Tensor, *, spatial_dim: Dim, collected_outputs: Optional[Dict[str, Tensor]] = None
    ) -> Tensor:
        """
        forward, single step or whole sequence.

        :param source: labels
        :param spatial_dim: single_step_dim or spatial dim of source
        :param collected_outputs:
        :return: final encoder output, after final layer norm
        """
        decoded = self.input_embedding(source) * self.input_embedding_scale
        if self.pos_enc is not None:
            decoded = decoded + self.pos_enc(spatial_dim=spatial_dim)
        decoded = rf.dropout(decoded, self.input_dropout)
        if self.input_embedding_proj is not None:
            decoded = self.input_embedding_proj(decoded)

        for layer_name, layer in self.layers.items():
            layer: TransformerEncoderLayer  # or similar
            decoded = layer(decoded, spatial_dim=spatial_dim)
            if collected_outputs is not None:
                collected_outputs[layer_name] = decoded

        decoded = self.final_layer_norm(decoded)
        return decoded


class TransformerEncoderLayer(rf.Module):
    """
    Represents a Transformer encoder block
    """

    def __init__(
        self,
        out_dim: Dim = Dim(512, name="transformer-enc-default-out-dim"),
        *,
        ff: Union[type, Dict[str, Any], rf.Module] = NotSpecified,
        dropout: float = 0.1,
        num_heads: int = 8,
        self_att: Optional[
            Union[rf.CausalSelfAttention, rf.RelPosCausalSelfAttention, rf.Module, type, Dict[str, Any]]
        ] = None,
        att_dropout: float = 0.1,
        norm: Union[type, Dict[str, Any], rf.Module, Callable] = rf.LayerNorm,
    ):
        """
        :param out_dim: the output feature dimension
        :param ff: feed-forward / MLP block. Default is :class:`FeedForward`
        :param dropout: the dropout value for the FF block
        :param num_heads: the number of attention heads
        :param self_att: the self-attention layer. SelfAttention originally and default
        :param att_dropout: attention dropout value
        :param norm: pre-normalization for FF and attention blocks
        """
        super().__init__()

        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()
        self.out_dim = out_dim

        if ff is NotSpecified:
            ff = FeedForward
        if isinstance(ff, rf.Module):
            ff = _copy.deepcopy(ff)
        else:
            ff_kwargs = dict(out_dim=out_dim, dropout=dropout)
            ff_kwargs = {k: v for (k, v) in ff_kwargs.items() if v is not NotSpecified}
            if isinstance(ff, type):
                ff = ff(**ff_kwargs)
            elif isinstance(ff, dict):
                ff = rf.build_from_dict(ff, **ff_kwargs)
            else:
                raise TypeError(f"unexpected ff type {ff!r}")
        assert isinstance(ff, rf.Module)

        self.ff = ff
        self.ff_layer_norm = make_norm(norm, out_dim)

        if self_att is None or isinstance(self_att, type) or isinstance(self_att, dict):
            self_att_opts_ = dict(
                in_dim=out_dim,
                proj_dim=out_dim,
                key_dim_total=out_dim,
                value_dim_total=out_dim,
                num_heads=num_heads,
                att_dropout=att_dropout,
            )
            if self_att is None:
                self.self_att = rf.SelfAttention(**self_att_opts_)
            elif isinstance(self_att, type):
                self.self_att = self_att(**self_att_opts_)
            elif isinstance(self_att, dict):
                self.self_att = rf.build_from_dict(self_att, **self_att_opts_)
            else:
                raise TypeError(f"unexpected self_att type {self_att!r}")
        elif isinstance(self_att, rf.Module):
            self.self_att = _copy.deepcopy(self_att)
        else:
            raise TypeError(f"unexpected self_att type {self_att!r}")
        self.self_att_layer_norm = make_norm(norm, out_dim)

    def default_initial_state(self, *, batch_dims: Sequence[Dim]) -> rf.State:
        """default initial state"""
        return rf.State(self_att=self.self_att.default_initial_state(batch_dims=batch_dims))

    def __call__(self, x: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        # (multi-head) self-attention (MHSA or simply SA)
        x_sa_ln = self.self_att_layer_norm(x)
        x_sa = self.self_att(x_sa_ln, axis=spatial_dim)
        x_sa = rf.dropout(x_sa, self.dropout, axis=self.dropout_broadcast and self.out_dim)
        x = x_sa + x

        # feed-forward (FF)
        x_ff_ln = self.ff_layer_norm(x)
        x_ff = self.ff(x_ff_ln)
        x_ff = rf.dropout(x_ff, self.dropout, axis=self.dropout_broadcast and self.out_dim)
        x = x_ff + x

        return x



"""
Public backend-related functions.
"""

from __future__ import annotations

from . import _backend

# And some functions from the internal backend API.
from ._backend import (
    select_backend,
    get_selected_backend,
    select_backend_torch,
    select_backend_returnn_layers_tf,
    is_executing_eagerly,
)

__all__ = [
    "select_backend",
    "get_selected_backend",
    "select_backend_torch",
    "select_backend_returnn_layers_tf",
    "is_executing_eagerly",
    "is_backend_raw_tensor_dim_tag_independent",
]


def is_backend_raw_tensor_dim_tag_independent() -> bool:
    """
    :return: whether raw tensors of the backend are independent of :class:`Dim`
        (Usually yes, e.g. :class:`tf.Tensor` or :class:`torch.Tensor`,
        but the TF-layers backend is an exception.)
    """
    return _backend.global_backend.is_backend_raw_tensor_dim_tag_independent



"""
Mel filterbank.
"""


from __future__ import annotations
from typing import Optional, Union, Tuple
import functools
import math
import numpy
from returnn.util import math as util_math
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["mel_filterbank", "log_mel_filterbank_from_raw"]


def mel_filterbank(
    x: Tensor,
    *,
    in_dim: Dim,
    out_dim: Dim,
    sampling_rate: Union[int, float],
    fft_length: Optional[int] = None,
    f_min: Optional[Union[int, float]] = None,
    f_max: Optional[Union[int, float]] = None,
):
    """
    Applies the Mel filterbank to the input.

    :param x:
    :param in_dim: expected to be fft_length // 2 + 1. E.g. via :func:`stft`.
    :param out_dim: nr of mel filters.
    :param sampling_rate:
    :param fft_length: fft_size, n_fft. Should match fft_length from :func:`stft`.
        If not given, infer this from in_dim, as (in_dim - 1) * 2.
    :param f_min:
    :param f_max:
    :return:
    """
    if not fft_length:
        fft_length = (in_dim.dimension - 1) * 2
    f_min = f_min or 0
    f_max = f_max or sampling_rate / 2.0
    # noinspection PyProtectedMember
    backend = x._raw_backend
    cache_key = (f_min, f_max, sampling_rate, fft_length, out_dim.dimension, backend, x.dtype, x.device)
    if cache_key in _mel_filter_bank_matrix_cache:
        filter_bank_matrix_: Tensor = _mel_filter_bank_matrix_cache[cache_key]
        # The in_dim might be temporarily created, so replace it by the new one.
        # Even allow for temporary out_dim.
        filter_bank_matrix = Tensor("mel-filter-bank", dims=(in_dim, out_dim), dtype=x.dtype)
        filter_bank_matrix.raw_tensor = filter_bank_matrix_.raw_tensor
    else:
        filter_bank_matrix_np = _mel_filter_bank_matrix_np(
            f_min=f_min, f_max=f_max, sampling_rate=sampling_rate, fft_size=fft_length, nr_of_filters=out_dim.dimension
        )
        filter_bank_matrix_np = filter_bank_matrix_np.astype(x.dtype)
        filter_bank_matrix = rf.convert_to_tensor(filter_bank_matrix_np, dims=(in_dim, out_dim), _backend=backend)
        filter_bank_matrix = rf.copy_to_device(filter_bank_matrix, x.device)
        if backend.executing_eagerly():
            if len(_mel_filter_bank_matrix_cache) > 100:
                # Very simple cache management. No LRU logic or anything like that.
                _mel_filter_bank_matrix_cache.clear()
            _mel_filter_bank_matrix_cache[cache_key] = filter_bank_matrix
    out = rf.matmul(x, filter_bank_matrix, reduce=in_dim)
    out.feature_dim = out_dim
    return out


# Used for eager frameworks. See mel_filterbank.
_mel_filter_bank_matrix_cache = {}


@functools.lru_cache()
def _mel_filter_bank_matrix_np(
    *,
    f_min: Union[int, float],
    f_max: Union[int, float],
    sampling_rate: Union[int, float],
    fft_size: int,
    nr_of_filters: int,
) -> numpy.ndarray:
    """
    Returns the filter matrix which yields the mel filter bank features, when applied to the spectrum as
    matmul(freqDom, filterMatrix), where freqDom has dimension (time, frequency)
    and filterMatrix is the matrix returned
    by this function.
    The filter matrix is computed according to equation 6.141 in
    [Huang & Acero+, 2001] "Spoken Language Processing - A Guide to Theroy, Algorithm, and System Development"

    :param float|int f_min: minimum frequency
    :param float|int f_max: maximum frequency
    :param float sampling_rate: sampling rate of audio signal
    :param int fft_size: dimension of discrete fourier transformation
    :param int nr_of_filters: number of mel frequency filter banks to be created
    :return: shape=(fft_size // 2 + 1, nr_of_filters), matrix yielding the mel frequency cepstral coefficients
    """

    def mel_scale(freq):
        """
        returns the respective value on the mel scale

        :param float freq: frequency value to transform onto mel scale
        :rtype: float
        """
        return 1125.0 * numpy.log(1 + float(freq) / 700)

    def inv_mel_scale(mel_val):
        """
        returns the respective value in the frequency domain

        :param float mel_val: value in mel domain
        :rtype: float
        """
        return 700.0 * (numpy.exp(float(mel_val) / 1125) - 1)

    # noinspection PyShadowingNames
    def filter_center(filter_id, f_min, f_max, sampling_rate, fft_size, nr_of_filters):
        """
        :param int filter_id: filter to compute the center frequency for
        :param float|int f_min: minimum frequency
        :param float|int f_max: maximum frequency
        :param float|int sampling_rate: sampling rate of audio signal
        :param int fft_size: dimension of discrete fourier transformation
        :param int nr_of_filters: number of mel frequency filter banks to be created
        :rtype: float
        :return: center frequency of filter
        """
        return (float(fft_size) / sampling_rate) * inv_mel_scale(
            mel_scale(f_min) + filter_id * ((mel_scale(f_max) - mel_scale(f_min)) / (nr_of_filters + 1))
        )

    filt_cent = numpy.zeros(shape=(nr_of_filters + 2,), dtype=numpy.float32)
    for i1 in range(nr_of_filters + 2):
        filt_cent[i1] = filter_center(i1, f_min, f_max, sampling_rate, fft_size, nr_of_filters)
    f_mat = numpy.zeros(shape=(fft_size // 2 + 1, nr_of_filters))
    for i1 in range(f_mat.shape[0]):
        for i2 in range(1, nr_of_filters + 1):
            if (i1 > filt_cent[i2 - 1]) and (i1 < filt_cent[i2 + 1]):
                if i1 < filt_cent[i2]:
                    num = i1 - filt_cent[i2 - 1]
                    denom = filt_cent[i2] - filt_cent[i2 - 1]
                else:
                    num = filt_cent[i2 + 1] - i1
                    denom = filt_cent[i2 + 1] - filt_cent[i2]
                el_val = num / denom
            else:
                el_val = 0
            f_mat[i1, i2 - 1] = el_val

    return f_mat


def log_mel_filterbank_from_raw(
    raw_audio: Tensor,
    *,
    in_spatial_dim: Dim,
    out_dim: Dim,
    sampling_rate: int = 16_000,
    window_len: float = 0.025,
    step_len: float = 0.010,
    n_fft: Optional[int] = None,
    log_base: Union[int, float] = 10,
) -> Tuple[Tensor, Dim]:
    """
    log mel filterbank features

    :param raw_audio: (..., in_spatial_dim, ...). if it has a feature_dim with dimension 1, it is squeezed away.
    :param in_spatial_dim:
    :param out_dim: nr of mel filters.
    :param sampling_rate: samples per second
    :param window_len: in seconds
    :param step_len: in seconds
    :param n_fft: fft_size, n_fft. Should match fft_length from :func:`stft`.
        If not provided, next power-of-two from window_num_frames.
    :param log_base: e.g. 10 or math.e
    """
    if raw_audio.feature_dim and raw_audio.feature_dim.dimension == 1:
        raw_audio = rf.squeeze(raw_audio, axis=raw_audio.feature_dim)
    window_num_frames = int(window_len * sampling_rate)
    step_num_frames = int(step_len * sampling_rate)
    if not n_fft:
        n_fft = util_math.next_power_of_two(window_num_frames)
    spectrogram, out_spatial_dim, in_dim_ = rf.stft(
        raw_audio,
        in_spatial_dim=in_spatial_dim,
        frame_step=step_num_frames,
        frame_length=window_num_frames,
        fft_length=n_fft,
    )
    power_spectrogram = rf.abs(spectrogram) ** 2.0
    mel_fbank = mel_filterbank(power_spectrogram, in_dim=in_dim_, out_dim=out_dim, sampling_rate=sampling_rate)
    log_mel_fbank = rf.safe_log(mel_fbank, eps=1e-10)
    if log_base != math.e:
        log_mel_fbank = log_mel_fbank * (1.0 / math.log(log_base))
    return log_mel_fbank, out_spatial_dim



"""
Helper for any type of PyTorch exceptions
"""

from __future__ import annotations

from typing import Optional, Union, Tuple

import torch
import numpy as np

from returnn.log import log
import returnn.frontend as rf
from returnn.tensor import TensorDict


def help_on_torch_exception(
    exc: Exception,
    *,
    step_idx: Optional[int] = None,
    extern_data: Optional[TensorDict] = None,
    model: Union[rf.Module, torch.nn.Module],
    always_direct_print: bool = False,
):
    """
    Gather some information which might be helpful for debugging a PyTorch exception.
    """
    from returnn.util.better_exchook import get_func_from_code_object, iter_traceback

    print(f"{type(exc).__name__}: {exc}", file=log.v1)

    exc_ext = [f"Step idx: {step_idx}"]
    if extern_data:
        exc_ext.append("Extern data:")
        if "seq_tag" in extern_data:
            exc_ext.append(f"  Seq tags: {extern_data['seq_tag'].raw_tensor}")
        covered_dim_tags = set()
        for data_key, data in extern_data.data.items():
            info, v_minmax = _help_data_or_array(data.raw_tensor)
            exc_ext.append(f"  {data_key}: {info}, {data}")
            if data.sparse:
                if v_minmax[0] < 0 or v_minmax[1] >= data.dim:
                    exc_ext.append(f"  WARNING, invalid label for data sparse dim {data.sparse_dim}")
            for dim in data.dims:
                if dim in covered_dim_tags:
                    continue
                covered_dim_tags.add(dim)
                if not dim.dyn_size_ext:
                    continue
                info, _ = _help_data_or_array(dim.dyn_size_ext.raw_tensor)
                exc_ext.append(f"    dim {dim.short_repr()} size: {info}")

    # Extend exception message by module call stack.
    exc_ext.append("Module call stack:")
    module_names_by_id = {}  # id -> name
    count_frames = 0
    for name, mod in model.named_modules():
        if id(mod) not in module_names_by_id:
            module_names_by_id[id(mod)] = name or "(root)"
    for frame in iter_traceback(exc.__traceback__):
        if frame.f_code.co_nlocals == 0:
            continue
        frame_self = frame.f_locals.get("self")
        if isinstance(frame_self, (torch.nn.Module, rf.Module)):
            func = get_func_from_code_object(frame.f_code, frame=frame)
            if func and func.__name__ and func.__name__.startswith("_") and not func.__name__.startswith("__"):
                continue
            func_name = (func and func.__qualname__) or type(frame_self).__name__
            exc_ext.append(f"({func_name}) {module_names_by_id.get(id(frame_self), '(unknown)')}")
            count_frames += 1
    if not count_frames:
        exc_ext.append("(No module call frames.)")

    if len(exc.args) == 1 and isinstance(exc.args[0], str) and not always_direct_print:
        exc.args = ("\n".join([exc.args[0], ""] + exc_ext),)
    else:
        for msg in exc_ext:
            print(msg, file=log.v3)


def _help_data_or_array(
    value: Union[torch.Tensor, np.ndarray, bool, object]
) -> Tuple[str, Tuple[Union[int, float], Union[int, float]]]:
    """
    :param value:
    :return: (info,(min,max))
    """
    if isinstance(value, torch.Tensor):
        value = value.detach().cpu().numpy()
    v_minmax = -1, -1
    if isinstance(value, np.ndarray):
        info = "shape %s, dtype %s" % (value.shape, value.dtype)
        if value.dtype.kind in "biuf":
            if value.size > 1:
                v_minmax = np.min(value), np.max(value)
                info += ", min/max %s/%s" % v_minmax
                if value.dtype.kind == "f":
                    info += ", mean/stddev %s/%s" % (np.mean(value), np.std(value))
                if value.ndim <= 1:
                    info += " (%s)" % np.array2string(value)
            elif value.size == 1:
                info += " (%s)" % np.array2string(value)
            else:
                info += ", EMPTY"
    elif isinstance(value, (np.floating, np.integer, np.bool_, float, int, bool, str, bytes)):
        info = "%s(%s)" % (type(value).__name__, value)
    elif value is None:
        info = "None"
    else:
        info = "type %r" % type(value)
    return info, v_minmax
